/*  dvdisaster: Additional error correction for optical media.
 *  Copyright (C) 2004,2005 Carsten Gnoerlich.
 *  Project home page: http://www.dvdisaster.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA,
 *  or direct your browser at http://www.gnu.org.
 */

#include "dvdisaster.h"

/***
 *** Print info about the prefix.* files
 ***/

void Print(char *imagename, char *eccname)
{  char adigest[33]; 
   struct stat mystat;
   FILE *ecc;
   EccHeader eh;
   ImageInfo *ii = NULL;
   gint64 excess_sectors = 0;
   gint64 eh_sectors = 0;
   int n;

   /*** See if we can access the .ecc file */

   ecc = fopen(eccname, "r");
#ifdef SYS_MINGW
   if(ecc) setmode(fileno(ecc), O_BINARY);
#endif

   if(ecc)
   {  n = fread(&eh, sizeof(EccHeader), 1, ecc);
      if(n && (!strncmp(eh.cookie,"*dvdisaster*",12)))
     	 eh_sectors = uchar_to_gint64(eh.sectors);
      else 
      {  fclose(ecc);
         ecc = NULL;
      }
   }

   /*** Examine the .img file */

   PrintMsg("\n%s: ", imagename);

   if(stat(imagename, &mystat) == -1)
   {  PrintMsg(_("not present\n"));
   }
   else
   {  gint64 sectors;
      int in_last;

      CalcSectors(mystat.st_size, &sectors, &in_last);
      PrintMsg(_("present, contains %lld medium sectors.\n"),sectors);

      if(ecc)
           ii = OpenImage(imagename, OPEN_FULLSCAN | VERBOSE_FULLSCAN | CHECK_CRC32, ecc);
      else ii = OpenImage(imagename, OPEN_FULLSCAN | VERBOSE_FULLSCAN, NULL);

      if(ii)  /* Compatibility with version 0.60 and beyond */
	CalculateFootprint(ii, &eh);

      /*** Peek into the ecc file to get expected sector count */

      if(ecc)  
      {  gint64 diff = 0;

	 if(ii->sectors < eh_sectors)
	 {  diff = eh_sectors - ii->sectors;

	    PrintMsg(_("* truncated image  : %lld sectors too short\n"), diff);
	    ii->sectorsMissing += diff;
	 }
	 if(ii->sectors > eh_sectors+2)
	 {  excess_sectors = ii->sectors - eh_sectors;
	 }
      }

      /*** Show summary of image read */

      if(excess_sectors)
      {  PrintMsg(_("* image too long   : %lld excess sectors\n"), excess_sectors);
      } 
      else
      {  if(!ii->sectorsMissing)
         {  
	    AsciiDigest(adigest, ii->mediumSum);
      
            if(!ii->crcErrors)
	       PrintMsg(_("- good image       : all sectors present\n"
		          "- image md5sum     : %s\n"),adigest);
	    else
               PrintMsg(_("* suspicious image : all sectors present, but %lld CRC errors\n"
		          "- image md5sum     : %s\n"),ii->crcErrors,adigest);

         }
         else
         {  if(!ii->crcErrors)
	         PrintMsg(_("* BAD image        : %lld sectors missing\n"), ii->sectorsMissing);
	    else PrintMsg(_("* BAD image        : %lld sectors missing, %lld CRC errors\n"), 
		         ii->sectorsMissing, ii->crcErrors);
         }
      }
   }

   /*** The .ecc file */

   PrintMsg("\n%s: ", eccname);

   if(stat(eccname, &mystat) == -1)
   {  PrintMsg(_("not present\n"));
   }
   else
   {  char method[5];
      gint64 ecc_blocks,ecc_expected,count;
      struct MD5Context md5ctxt;
      char buf[1024];
      unsigned char digest[16];
      int percent,last_percent;

      if(!ecc)
      {  PrintMsg(_(".ecc file %s invalid or unreadable.\n"), eccname);
         goto print_end;
      }
      
      strncpy(method, eh.method, 4); method[4] = 0;

      if(!eh.neededVersion)  /* The V0.41.* series did not fill in this field. */
	eh.neededVersion = 4100;
      
      if(eh.creatorVersion)
      {  if(eh.creatorVersion%100)        
	   PrintMsg(_("created by dvdisaster-%d.%d.%d.\n"), 
		    eh.creatorVersion/10000,
		    (eh.creatorVersion%10000)/100,
		    eh.creatorVersion%100);
	 else
	   PrintMsg(_("created by dvdisaster-%d.%d\n"), 
		    eh.creatorVersion/10000,
		    (eh.creatorVersion%10000)/100);
      }
      else
	PrintMsg(_("created by dvdisaster-0.41.x.\n"));

      PrintMsg(_("- method           : %4s, %d roots, %4.1f%% redundancy.\n"),
	       method, eh.eccBytes, ((double)eh.eccBytes*100.0)/(double)(255-eh.eccBytes));

      if(!VerifyVersion(&eh, 0))
	PrintMsg(_("- requires         : dvdisaster-%d.%d (good)\n"),
		 eh.neededVersion/10000,
		 (eh.neededVersion%10000)/100);
      else 
	PrintMsg(_("* requires         : dvdisaster-%d.%d (BAD)\n"
		   "* Warning          : The following output might be incorrect.\n"
		   "*                  : Please visit http://www.dvdisaster.com for an upgrade.\n"),
		 eh.neededVersion/10000,
		 (eh.neededVersion%10000)/100);

      eh_sectors = uchar_to_gint64(eh.sectors);
      if(ii)
      {  if(ii->sectors == eh_sectors)
	       PrintMsg(_("- medium sectors   : %lld (good)\n"),eh_sectors);
         else 
	 { if(ii->sectors > eh_sectors && ii->sectors - eh_sectors <= 2)   
	        PrintMsg(_("* medium sectors   : %lld (BAD, perhaps TAO/DAO mismatch)\n"),eh_sectors);
           else PrintMsg(_("* medium sectors   : %lld (BAD)\n"),eh_sectors);
	 }
      }
      else      PrintMsg(_("- medium sectors   : %lld\n"),eh_sectors);

      /*** Verify md5sums against image and map (if present) */

      AsciiDigest(adigest, eh.mediumSum);
      if(ii && !ii->sectorsMissing && !excess_sectors)
      {  n = !memcmp(eh.mediumSum, ii->mediumSum, 16);
         if(n) PrintMsg(_("- image md5sum     : %s (good)\n"),adigest);
	 else  PrintMsg(_("* image md5sum     : %s (BAD)\n"),adigest);
      }
      else PrintMsg(_("- image md5sum     : %s\n"),adigest);

      if(ii)
      {  if(!memcmp(ii->mediumFP, "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000", 16))
	      PrintMsg(_("* fingerprint match: NOT POSSIBLE - sector 0 is missing in image!\n"));
         else
	 if(memcmp(ii->mediumFP, eh.mediumFP, 16)) 
	      PrintMsg(_("* fingerprint match: MISMATCH - .img and .ecc don't belong together!\n"));
         else PrintMsg(_("- fingerprint match: good\n"));
      }

      ecc_expected = 2048*((eh_sectors+eh.dataBytes-1)/eh.dataBytes);
      ecc_blocks = (mystat.st_size-eh_sectors*sizeof(guint32)-sizeof(EccHeader))/eh.eccBytes;

      if(ecc_expected == ecc_blocks)
	   PrintMsg(_("- ecc blocks       : %lld (good)\n"),ecc_blocks);
      else PrintMsg(_("* ecc blocks       : %lld (BAD, expected %lld)\n"),ecc_blocks,ecc_expected);

      /*** Test ecc file against its own md5sum */

      MD5Init(&md5ctxt);

      last_percent = -1;
      count = sizeof(EccHeader);

      if(fseeko(ecc, (gint64)sizeof(EccHeader), SEEK_SET) == -1)
	Stop(_("fseek failed skipping ecc header: %s"),strerror(errno));

      while(!feof(ecc))
      {  n = fread(buf, 1, 1024, ecc);
         MD5Update(&md5ctxt, buf, n);

	 count += n;
	 percent = (100*count)/mystat.st_size;
	 if(last_percent != percent) 
	 {  PrintMsg(_("- ecc md5sum       : %3d%%\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"),percent);
	    last_percent = percent;
	 }

      }
      MD5Final(digest, &md5ctxt);
      AsciiDigest(adigest, digest);

      if(memcmp(eh.eccSum, digest, 16))
	   PrintMsg(_("* ecc md5sum       : BAD, ecc file may be damaged!\n"));
      else PrintMsg(_("- ecc md5sum       : %s (good)\n"),adigest);
   }

print_end:
   PrintMsg("\n");
   if(ecc) fclose(ecc);
   if(ii) FreeImageInfo(ii);
}
